# Plain-text shell archive construction tool
# J. Welsh, 2017
#
# Warnings to archive authors:
# - Absolute paths in archives are evil. Don't override the checks.
# - The code is believed to safely handle the gamut of evil shell characters in
#   paths (spaces, quotes, leading dashes, newlines...). You are still strongly
#   discouraged from using these.
# - To ensure portable, predictable, perhaps even parseable archives, avoid
#   emitting code directly.
# - Encode any blobs ("F", "X" commands, "a" if unsure).
#
# To avoid executing evil archives, treat them as any other shell script:
# - Verify signatures and mind your Web of Trust.
# - The generated archives are designed to be legible. Read!

cat <<'.'
#!/bin/sh
# Self-extracting archive generated by ar.sh (gales-util 97K)
dir=${1:-.}
printf 'Extracting to %s\n' "$dir"
cd -- "$dir" || exit
umask 022
if [ `id -u` -eq 0 ]; then
	attrs () { chown -- "$3" "$1"; chmod -- "$2" "$1"; }
else
	attrs () { chmod -- "$2" "$1"; }
fi
_rm () { rm -f -- "$@"; }
_mv () { mv -f -- "$@"; }
ln_s () { ln -s -- "$@"; }
mkdir_p () { mkdir -p -- "$@"; }
txt () { awk 'NR>1{print l} {l=substr($0,2)} END{ORS="";print l}'; }
bin () { base64 -d; }
d () {
	printf %s/\\n "$1"
	mkdir_p "$1" || return
	attrs "$@"
}
unpack () {
	mode=$1; shift
	f=$1; shift
	d=`dirname -- "$f"`
	t="$f.__new"
	printf %s\\n "$f"
	mkdir_p "$d" || return
	_rm "$t" || return
	$mode >"$t" || return
	attrs "$t" "$@"
	_mv "$t" "$f"
}
f () { unpack txt "$@"; }
F () { unpack bin "$@"; }
l () {
	src=$1; dst=$2
	d=`dirname -- "$src"`
	printf '%s -> %s\n' "$src" "$dst"
	if [ -d "$src" ]; then
		mkdir_p "$d/__new" || return
		f=`basename -- "$src"`
		t="$d/__new/$f"
		_rm "$t" || return
		ln_s "$dst" "$t" || return
		_mv "$t" "$d" || _rm "$t"
		rmdir -- "$d/__new"
	else
		mkdir_p "$d" || return
		t="$src.__new"
		_rm "$t" || return
		ln_s "$dst" "$t" || return
		_mv "$t" "$src"
	fi
}
p () {
	d=`dirname -- "$1"`
	printf %s\\n "$1"
	mkdir_p "$d" || return
	_rm "$1" || return
	mkfifo -- "$1" || return
	attrs "$@"
}
i () { printf '>> INFO: %s\n' "$1"; }
#end-preamble
.

check_abs () {
	if [ `expr "$1" : /` != 0 ]; then
		printf 'absolute paths not allowed: %s\n' "$1" >&2
		return 1
	fi
}
# from Rich Felker
quote () { printf %s\\n "$1" | sed "s/'/'\\\\''/g;1s/^/'/;\$s/\$/'/"; }

# directory
d () {
	check_abs "$1" || return 1
	printf 'd %s %s %s\n' "`quote "$1"`" "${2:-755}" "${3:-0:0}"
}
# file
f () {
	check_abs "$1" || return 1
	src=${2:-`basename -- "$1"`}
	printf "f %s %s %s <<'.'"\\n "`quote "$1"`" "${3:-644}" "${4:-0:0}"
	# blockquote in case file contains the delimiter
	# ("paste" would fit better here but busybox lacks it)
	sed 's/^/	/' <"$src" # literal tab char (\t is nonstandard)
	# trailing NL in case file lacks
	printf '\n.\n'
	# the extra whitespace is removed by the "txt" awkism above, which
	# unfortunately requires linear memory; head -c -1 is nonstandard
}
# binary file
F () {
	check_abs "$1" || return 1
	src=${2:-`basename -- "$1"`}
	printf "F %s %s %s <<'.'"\\n "`quote "$1"`" "${3:-644}" "${4:-0:0}"
	base64 <"$src"
	echo .
}
# executable
x () { f "$1" "$2" "${3:-755}" "$4"; }
# binary executable
X () { F "$1" "$2" "${3:-755}" "$4"; }
# symlink
l () {
	check_abs "$1" || return 1
	printf 'l %s %s\n' "`quote "$1"`" "`quote "$2"`"
}
# pipe
p () {
	check_abs "$1" || return 1
	printf 'p %s %s %s\n' "`quote "$1"`" "${2:-644}" "${3:-0:0}"
}
# autodetect (recursive, excluding dotfiles)
a () {
	check_abs "$1" || return 1
	src=${2:-`basename -- "$1"`}
	if [ -L "$src" ]; then
		l "$1" "`readlink -n -- "$src"`"
	elif [ ! -e "$src" ]; then
		printf 'no such file: %s\n' "$src" >&2
		return 1
	elif [ -f "$src" ]; then
		if [ `LC_CTYPE=C tr -d '[:alnum:][:punct:][:space:]' <"$src" |
			dd bs=1 count=1 2>/dev/null | wc -c` -ne 0 ]; then
			# ^ busybox tr is missing print/graph classes
			if [ -x "$src" ]
			then X "$1" "$src"
			else F "$1" "$src"; fi
		else
			if [ -x "$src" ]
			then x "$1" "$src"
			else f "$1" "$src"; fi
		fi
	elif [ -d "$src" ]; then
		d "$1"
		cd -- "$src"
		for f in *; do [ -L "$f" -o -e "$f" ] && a "$1/$f" "$f"; done
		cd ..
	elif [ -p "$src" ]; then
		p "$1"
	else
		printf 'unsupported file type: %s\n' "$1" >&2
		return 1
	fi
}
# informational message
i () { printf 'i %s\n' "`quote "$1"`"; }
